#!/usr/bin/env python
import sys
import os
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import numpy as np
from csv import DictReader
from copy import copy
from collections import defaultdict
from matplotlib.ticker import FuncFormatter

#regular
LINE_STYLE = ['b:+', 'g-', 'r-s', 'c--', 'k-.', '--o']
#increasing
#LINE_STYLE = ['b--s', 'b-s', 'g--x', 'g-x', 'r--', 'r-', 'k-.']
#reqnum
#LINE_STYLE = ['b-.', 'r--', 'g-', 'c:+', 'k:']

schednames  = ['EDF', 'N-PST', 'CBA-PST', 'WBA-PST', 'N-RST', 'CBA-RST', 'WBA-RST', 'ALL']
legendnames = ['EDF', 'N-PST', 'CBA-PST', 'WBA-PST', 'N-RST', 'CBA-RST', 'WBA-RST', 'ALL']

def to_percent(y, position):
    # Ignore the passed in position. This has the effect of scaling the default
    # tick locations.
    s = str(int((10000*y/756.0)+0.5)/100)

    # The percent symbol needs escaping in latex
    if matplotlib.rcParams['text.usetex'] is True:
        return s + r'$\%$'
    else:
        return s + '%'
        
def main():
    datadir = "../results_refined/"
    
    total_count = 0

    count = 0.0
    tots_reg = {'CBA-PST' : 0, 'CBA-RST' : 0, 'WBA-PST' : 0, 'WBA-RST' : 0} 
    tots_way = {'CBA-PST' : 0, 'CBA-RST' : 0, 'WBA-PST' : 0, 'WBA-RST' : 0} 
    ntots = {'CBA-PST' : 0, 'CBA-RST' : 0, 'WBA-PST' : 0, 'WBA-RST' : 0} 
    
    totsch = {'CBA-PST' : 0, 'CBA-RST' : 0, 'WBA-PST' : 0, 'WBA-RST' : 0, 'N-PST' : 0, 'N-RST' : 0} 
    
    #PST v RTP, percent increase in sched under PST
    svr_perc = {'CBA' : 0, 'WBA' : 0, 'N' : 0, 'CBA-C' : 0, 'WBA-C' : 0, 'N-C' : 0} 
    
    all_perc = 0
    
    perc_of_all = {'CBA-PST' : 0, 'CBA-RST' : 0, 'WBA-PST' : 0, 'WBA-RST' : 0, 'N-PST' : 0, 'N-RST' : 0} 
    
    ntots_count_r = 0
    ntots_count_p = 0
    max_count = 0
    
    chcount = 0
    abcount = 0
    lucount = 0
    hucount = 0
    spcount = 0
    lpcount = 0
    hdcount = 0
    hocount = 0
    
    hoarea = {'CBA-PST' : 0, 'CBA-RST' : 0, 'WBA-PST' : 0, 'WBA-RST' : 0, 'N-PST' : 0, 'N-RST' : 0} 
    loarea = {'CBA-PST' : 0, 'CBA-RST' : 0, 'WBA-PST' : 0, 'WBA-RST' : 0, 'N-PST' : 0, 'N-RST' : 0} 
    
    tcount = 0
    
    lowcount = 0
    for filename in os.listdir(os.getcwd() + "/" + datadir):
        d = DictReader(open(datadir + filename, 'r'))
        data = defaultdict(list)
        for row in d:
            for key, value in row.iteritems():
                data[key].append(value)
        
        npdata = defaultdict(list)
        npdata["sys_util"] = np.asarray(data["sys_util"])
        idx_ord = npdata["sys_util"].argsort()
    
        for key in data.keys():
            npdata[key] = np.asarray(data[key])[idx_ord]
    
        areas = defaultdict()
        for sch_i, sched in enumerate(schednames):
            areas[sched] = 0
            for dp_i in range(1, len(npdata[sched])):
                areas[sched] += (float(npdata[sched][dp_i])+float(npdata[sched][dp_i-1]))*0.2
    
        if areas['N-PST'] == 0:
            areas['N-PST'] = .2
        for dp_i in range(len(npdata['Count'])):
            max_count = max(max_count, int(npdata['Count'][dp_i]))
    
        if 'regular' in filename:
            tots_reg['WBA-PST'] += areas['WBA-PST'] - areas['CBA-PST']
            tots_reg['WBA-RST'] += areas['WBA-RST'] - areas['CBA-RST']
            tots_reg['CBA-PST'] += areas['CBA-PST'] - areas['WBA-PST']
            tots_reg['CBA-RST'] += areas['CBA-RST'] - areas['WBA-RST']
        else:
            tots_way['WBA-PST'] += areas['WBA-PST'] - areas['CBA-PST']
            tots_way['WBA-RST'] += areas['WBA-RST'] - areas['CBA-RST']
            tots_way['CBA-PST'] += areas['CBA-PST'] - areas['WBA-PST']
            tots_way['CBA-RST'] += areas['CBA-RST'] - areas['WBA-RST']

        if areas['N-PST'] != 0:
            ntots['CBA-PST'] += 100*((areas['CBA-PST'] - areas['N-PST'])/areas['N-PST'])
            ntots['WBA-PST'] += 100*((areas['WBA-PST'] - areas['N-PST'])/areas['N-PST'])
            ntots_count_p += 1
        if areas['N-RST'] != 0:
            ntots['CBA-RST'] += 100*((areas['CBA-RST'] - areas['N-RST'])/areas['N-RST'])
            ntots['WBA-RST'] += 100*((areas['WBA-RST'] - areas['N-RST'])/areas['N-RST'])
            ntots_count_r += 1
    
        if areas['CBA-RST'] != 0:
            svr_perc['CBA'] += 100*((areas['CBA-PST']/areas['CBA-RST'])-1)
            svr_perc['CBA-C'] += 1
        if areas['WBA-RST'] != 0:
            svr_perc['WBA'] += 100*((areas['WBA-PST']/areas['WBA-RST'])-1)
            svr_perc['WBA-C'] += 1
        if areas['N-RST'] != 0:
            svr_perc['N'] += 100*((areas['N-PST']/areas['N-RST'])-1)
            svr_perc['N-C'] += 1
    
        best_curve = max(areas['N-PST'], areas['CBA-PST'], areas['WBA-PST'], areas['N-RST'], areas['CBA-RST'], areas['WBA-RST'])
        all_perc += 100*((areas['ALL']/best_curve)-1)
        
        for scheme in ['CBA-PST', 'CBA-RST', 'WBA-PST', 'WBA-RST', 'N-PST', 'N-RST']:
            perc_of_all[scheme] += 100*(areas[scheme]/areas['ALL'])
        
        if areas['CBA-PST'] - areas['N-PST'] > 0.5 and areas['N-PST'] - areas['N-RST'] > 0.0 and areas['ALL'] > 2 and npdata["pet_sens"][0] == 'way-heavy':
            print areas['N-PST'] - areas['N-RST'], filename
            
#         if areas['N-PST'] - areas['N-RST'] > 0.1 and areas['ALL'] > 4 and npdata['mode_count'][0] == 'uni-large':
#             print areas['N-PST'] - areas['N-RST'], filename
        for scheme in ['CBA-PST', 'CBA-RST', 'WBA-PST', 'WBA-RST', 'N-PST', 'N-RST']:
            totsch[scheme] += areas[scheme]
        
        
        if areas['N-PST'] != 0 and 100*((areas['CBA-PST'] - areas['N-PST'])/areas['N-PST']) <= 10:
            lowcount += 1
            if 'light-light-heavy' in filename:
                chcount += 1
            else:
                abcount += 1
            if npdata['task_util'][0] == 'uni-light':
                lucount += 1
            else:
                hucount += 1
            if npdata['period'][0] == 'harmonic-short':
                spcount += 1
            else:
                lpcount += 1
            if npdata['dram'][0] == 'uni-heavy':
                hdcount += 1
            if npdata['mode_olap'][0] == 'uni-light':
                hocount += 1
                
            if npdata['task_util'][0] == 'uni-heavy' or npdata['period'][0] == 'harmonic-short':
                tcount += 1
        
        if npdata['mode_olap'][0] == 'uni-heavy':
            for scheme in ['CBA-PST', 'CBA-RST', 'WBA-PST', 'WBA-RST', 'N-PST', 'N-RST']:
                hoarea[scheme] += areas[scheme]
        if npdata['mode_olap'][0] == 'uni-light':
            for scheme in ['CBA-PST', 'CBA-RST', 'WBA-PST', 'WBA-RST', 'N-PST', 'N-RST']:
                loarea[scheme] += areas[scheme]
        
        count += 1.0
    
    print "util-gains reg:"
    for key in ['CBA-PST', 'CBA-RST', 'WBA-PST', 'WBA-RST']:
        print "", key, ":", float(tots_reg[key])/(count*0.5)
    print "util-gains way:"
    for key in ['CBA-PST', 'CBA-RST', 'WBA-PST', 'WBA-RST']:
        print "", key, ":", float(tots_way[key])/(count*0.5)
        
    print "average percent sched more PST over RST:"
    for key in ['CBA', 'WBA', 'N']:
        print "", key, ":", float(svr_perc[key])/float(svr_perc[key+'-C'])
        
    print "overall percent sched more PST over RST:"
    for key in ['CBA', 'WBA', 'N']:
        print "", key, ":", 100*(float(totsch[key+'-PST'])/float(totsch[key+'-RST'])-1)

    print "ALL is on average at least", all_perc/count, "% better than other schemes"
    
    print "Percent of all schedulability each scheme achieved"
    
    for scheme in ['CBA-PST', 'CBA-RST', 'WBA-PST', 'WBA-RST', 'N-PST', 'N-RST']:
        print scheme, ":", perc_of_all[scheme]/count
        
    

    print
    print
    print "percent gains over N:"
    for key in ['CBA-PST', 'WBA-PST']:
        print "", key, ":", float(ntots[key])/ntots_count_p
    for key in ['CBA-RST', 'WBA-RST']:
        print "", key, ":", float(ntots[key])/ntots_count_r
    print ntots_count_p, "and", ntots_count_r, "out of", count, "checked."
        
        
    print "percent over N for all scenarios combined:"
    for key in ['CBA-PST', 'WBA-PST']:
        print "", key, ":", 100*(float(totsch[key])/float(totsch['N-PST'])-1)
    for key in ['CBA-RST', 'WBA-RST']:
        print "", key, ":", 100*(float(totsch[key])/float(totsch['N-RST'])-1)

    print "t max count:", max_count
    
    print chcount, abcount
    print lucount, hucount
    print spcount, lpcount
    print tcount
    print hdcount
    print "total with small diff:", lowcount, "out of", count
    print "heavy mode overlap:", hocount
    
    print
    print "sched change moving to heavy overlap"
    for scheme in ['CBA-PST', 'CBA-RST', 'WBA-PST', 'WBA-RST', 'N-PST', 'N-RST']:
        print "", scheme, ":", (float(loarea[scheme]) - float(hoarea[scheme]))/float(loarea[scheme])
if __name__ == '__main__':
    main()