#!/usr/bin/python

from Tkinter import *
import ttk
import schedcat.model.LLCLayout as LLC
import time
from LLCOptimizer import *

class Custom:
    pass

class LLCCanvas:

    def __init__(self, scheme, tsMCX):
    
        self.tsMCX = tsMCX
        
        self.scheme = scheme
        
        root = Tk()
        root.title("Cache")
        root.resizable(0,0)
        
        canvFrame = Frame(root)
        canvFrame.pack(side = LEFT)
        self.canvas = Canvas(canvFrame, bg="white", width=800, height= 800)
        self.canvas.configure(cursor="crosshair")
        self.canvas.pack(side = LEFT)
        
        self.canvas.bind("<B1-Motion>", self.moveline)
        self.canvas.bind("<Button-1>", self.select)
        self.canvas.bind("<ButtonRelease-1>", self.deselect)
        
        self.draw_cache()
        
        menuFrame = Frame(root)
        menuFrame.pack(side = RIGHT)
        btn = Button(menuFrame, text="LP Opt Ways", command=self.LP_Opt_Ways)
        btn.pack (side = TOP)
        btn = Button(menuFrame, text="LP Opt Sets", command=self.LP_Opt_Sets)
        btn.pack (side = TOP)
        btn = Button(menuFrame, text="MILP Opt Ways", command=self.MILP_Opt_Ways)
        btn.pack (side = TOP)
        btn = Button(menuFrame, text="MILP Opt Sets", command=self.MILP_Opt_Sets)
        btn.pack (side = TOP)
        
        root.mainloop()
    
    def LP_Opt_Ways(self):
        newscheme = LPLLCways(self.tsMCX, 'Continuous', self.scheme, olap = True)
        if newscheme is None:
            newscheme = LPLLCways(self.tsMCX, 'Continuous', self.scheme, olap = False)
            if newscheme is None:
                return
        self.scheme = newscheme
        self.draw_cache()
        
    def LP_Opt_Sets(self):
        newscheme = LPLLCsets(self.tsMCX, 'Continuous', self.scheme)
        if newscheme is None:
            return
        self.scheme = newscheme
        self.draw_cache()
        
    def MILP_Opt_Ways(self):
        newscheme = LPLLCways(self.tsMCX, 'Integer', self.scheme, olap = True)
        if newscheme is None:
            newscheme = LPLLCways(self.tsMCX, 'Integer', self.scheme, olap = False)
            if newscheme is None:
                return
        self.scheme = newscheme
        self.draw_cache()
        
    def MILP_Opt_Sets(self):
        newscheme = LPLLCsets(self.tsMCX, 'Integer', self.scheme)
        if newscheme is None:
            return
        self.scheme = newscheme
        self.draw_cache()
    
    def canv_coords(self):
        A = self.scheme.A
        B = self.scheme.B
        C = self.scheme.C
        S = self.scheme.S
        rel_coords = Custom()
        m = self.scheme.m
        
        rel_coords.A = [self.X0+A[i]*self.uX for i in range(0,m)]
        rel_coords.B = [self.Xf-C*self.uX-B[i]*self.uX for i in range(0,m)]
        rel_coords.C = self.Xf-C*self.uX
        rel_coords.S = [self.Y0]
        rel_coords.S.extend([self.Y0+sum([S[i] for i in range(0,j)])*self.uY for j in range(1, m)])
        rel_coords.S.append(self.Yf)
        
        return rel_coords
    
    def LLC_coords(self, event):
        LLC_coord = Custom()
        LLC_coord.W = (event.x-float(self.X0))/self.uX
        LLC_coord.S = (event.y-float(self.Y0))/self.uY
        return LLC_coord
    
    def LLC_size(self, LLC_coord, area):
        if area[0] is 'A':
            return LLC_coord.W
        if area[0] is 'B':
            return (self.scheme.mW-self.scheme.C)-LLC_coord.W
        if area[0] is 'C':
            return self.scheme.mW-LLC_coord.W
        if area[0] is 'S':
            return LLC_coord.S-sum([self.scheme.S[i] for i in range(0, int(area[1])-1)])
    
    def draw_cache(self):
    
        self.canvas.delete(ALL)
        
        border = 100
        self.X0 = border
        self.Y0 = border
        self.Xf = int(self.canvas.cget("width"))-border
        self.Yf = int(self.canvas.cget("height"))-border
        
        self.uX = float(self.Xf-self.X0)/float(self.scheme.mW)
        self.uY = float(self.Yf-self.Y0)/float(self.scheme.mS)
        
        self.canvas.create_rectangle(self.X0, self.Y0, self.Xf, self.Yf, tags="cache") 

        m = len(self.scheme.A)
        
        coords = self.canv_coords()
            
        for i in range(0, m):
            self.canvas.create_line(coords.A[i], coords.S[i], coords.A[i], coords.S[i+1], tags="A"+str(i), fill = "red")
            self.canvas.create_line(coords.B[i], coords.S[i], coords.B[i], coords.S[i+1], tags="B"+str(i), fill = "blue")
        for i in range(1, m):
            self.canvas.create_line(self.X0, coords.S[i], coords.C, coords.S[i], tags="S"+str(i), fill = "black")
        self.canvas.create_line(coords.C, self.Y0, coords.C, self.Yf, tags="C", fill="green")
        
        self.tsMCX.setScheme(self.scheme.rounded())
        
        txtmarg = 50
        txtline = 15
        lines = 0
        self.canvas.create_text(self.Xf+txtmarg, self.Y0 + txtline*lines,
                                text="UA_C : " + str(round(self.tsMCX.level('A').utilizationI('C'), 2)))
        lines += 1
        self.canvas.create_text(self.Xf+txtmarg, self.Y0 + txtline*lines,
                                text="UB_C : " + str(round(self.tsMCX.level('B').utilizationI('C'), 2)))
        lines += 1                        
        self.canvas.create_text(self.Xf+txtmarg, self.Y0 + txtline*lines,
                                text="UC_C : " + str(round(self.tsMCX.level('C').utilizationI('C'), 2)))
        lines += 1                        
        self.canvas.create_text(self.Xf+txtmarg, self.Y0 + txtline*lines,
                                text="U__C : " + str(round(self.tsMCX.utilizationI('C'), 2)))
                                
                                
        lines +=2
        self.canvas.create_text(self.Xf+txtmarg, self.Y0 + txtline*lines,
                                text="UA_B : " + str(round(self.tsMCX.level('A').utilizationI('B'), 2)))
        lines += 1
        self.canvas.create_text(self.Xf+txtmarg, self.Y0 + txtline*lines,
                                text="UB_B : " + str(round(self.tsMCX.level('B').utilizationI('B'), 2)))
        lines += 1
        self.canvas.create_text(self.Xf+txtmarg, self.Y0 + txtline*lines,
                                text="U__B : " + str(round(self.tsMCX.level('A').utilizationI('B')+self.tsMCX.level('B').utilizationI('B'), 2)))
                      
                      
        lines +=2
        self.canvas.create_text(self.Xf+txtmarg, self.Y0 + txtline*lines,
                                text="UA_A : " + str(round(self.tsMCX.level('A').utilizationI('A'), 2)))
                                
                                
        lines += 2   
        for i in range(0, self.scheme.m):
            lines += 1                        
            self.canvas.create_text(self.Xf+txtmarg, self.Y0 + txtline*lines,
                                    text="UA_A" + str(i) + ": " + str(round(self.tsMCX.level('A').core(i).utilizationI('A'), 2)))
        lines +=1
        for i in range(0, self.scheme.m):
            lines += 1                        
            self.canvas.create_text(self.Xf+txtmarg, self.Y0 + txtline*lines,
                                    text="UA_B" + str(i) + ": " + str(round(self.tsMCX.level('A').core(i).utilizationI('B'), 2)))
        lines +=1
        for i in range(0, self.scheme.m):
            lines += 1                        
            self.canvas.create_text(self.Xf+txtmarg, self.Y0 + txtline*lines,
                                    text="UA_C" + str(i) + ": " + str(round(self.tsMCX.level('A').core(i).utilizationI('C'), 2)))
                                    
        
        lines +=2
        for i in range(0, self.scheme.m):
            lines += 1                        
            self.canvas.create_text(self.Xf+txtmarg, self.Y0 + txtline*lines,
                                    text="UB_B" + str(i) + ": " + str(round(self.tsMCX.level('B').core(i).utilizationI('B'), 2)))
        lines +=1
        for i in range(0, self.scheme.m):
            lines += 1                        
            self.canvas.create_text(self.Xf+txtmarg, self.Y0 + txtline*lines,
                                    text="UB_C" + str(i) + ": " + str(round(self.tsMCX.level('B').core(i).utilizationI('C'), 2)))
                                    
        lines +=1
        for i in range(0, self.scheme.m):
            lines += 1                        
            self.canvas.create_text(self.Xf+txtmarg, self.Y0 + txtline*lines,
                                    text="U__B" + str(i) + ": " + str(round(self.tsMCX.level('A').core(i).utilizationI('B') + self.tsMCX.level('B').core(i).utilizationI('B'), 2)))
                                    
    
    def select(self, event):
        range = 7
        closest = self.canvas.find_closest(event.x, event.y, halo=None, start=None)
        in_range = self.canvas.find_overlapping(event.x-range, event.y-range, event.x+range, event.y+range)
        if closest[0] in in_range:
            self.selected = self.canvas.gettags(closest)[0]
        
    def deselect(self, event):
        self.selected = None
        
    def moveline(self, event):
        if not self.selected is None:
            LLC_coords = self.LLC_coords(event)
            new_val = self.LLC_size(LLC_coords, self.selected)
            
            old_val = self.scheme.get(self.selected)
            self.scheme.set(self.selected, new_val)
            
            
            if self.selected[0] is 'S':
                idxNext = int(self.selected[1])
                if not idxNext is 3:
                    self.scheme.S[idxNext] = self.scheme.S[idxNext]-(new_val-old_val)
            
            self.draw_cache()
