#!/usr/bin/env python

from collections import defaultdict

from distributions import *
from schedcat.model.LLCLayout import *
from schedcat.model.LLC_params import *
import schedcat.model.ohc as ohc
import random
import time

from pulp import *
import gurobipy

import numpy as np

# capped_wss = {'B' : [512, 512, 512, 512],
#               'C' : 512}

def LPLLCWays(ts_mcx, vartype, csch = bypass_scheme(), olap = True, debug = False, noIso = False):
    capped_wss = {'B' : [(0 if len(ts_mcx.level('B').core(c)) == 0 else max([t.wss for t in ts_mcx.level('B').core(c)])) for c in range(4)], 
                  'C' : (0 if len(ts_mcx.level('C')) == 0 else max([t.wss for t in ts_mcx.level('C')]))}
    ts_mcx.setScheme(csch)
    ts_mcx.A = ts_mcx.level('A')
    ts_mcx.B = ts_mcx.level('B')
    ts_mcx.C = ts_mcx.level('C')
     
    #----------------------- Gather Data -----------------------------#
    UA_Adata = np.zeros((CORES, csch.mW+1))
    UA_Bdata = np.zeros((CORES, csch.mW+1))
    UA_Cdata = np.zeros((CORES, csch.mW+1))
    
    UB_Bdata = np.zeros((CORES, csch.mW+1))
    UB_Cdata = np.zeros((CORES, csch.mW+1))
    
    UC_Cdata = np.zeros(csch.mW+1)
    
    hdata = np.zeros(csch.mW+1)
    Hdata = np.zeros(csch.mW+1)
    
    minAperiods = []
    for core in range(0, 4):
        if len(ts_mcx.A.core(core)) == 0:
            minAperiods.append(0)
        else:
            minAperiods.append(min([t.period for t in ts_mcx.A.core(core)]))
    
    for w in range(0,csch.mW+1):
        wsch = scheme([area(w, csch.A[i].S) for i in range(0, CORES)],
                      [area(w, csch.B[i].S) for i in range(0, CORES)],
                      area(w, csch.C.S), csch.mW, csch.mS)
                      
        if noIso:
            wsch = scheme([area(0, csch.A[i].S) for i in range(0, CORES)],
                          [area(0, csch.B[i].S) for i in range(0, CORES)],
                          area(0, csch.C.S), csch.mW, csch.mS)
                          
        ts_mcx.setScheme(wsch)
        for core in range(0, CORES):
            UA_Adata[core][w] = ts_mcx.A.core(core).utilization('A', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
            UA_Bdata[core][w] = ts_mcx.A.core(core).utilization('B', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
            UA_Cdata[core][w] = ts_mcx.A.core(core).utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
            
            UB_Bdata[core][w] = ts_mcx.B.core(core).utilization('B', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
            UB_Cdata[core][w] = ts_mcx.B.core(core).utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
        
        UAi = defaultdict()
        UAi['B'] = [UA_Bdata[core][w] for core in range(0, CORES)]
        UAi['C'] = [UA_Cdata[core][w] for core in range(0, CORES)]
        
        UAi = ts_mcx.secondaryInflateConst(UAi, minAperiods = minAperiods)
        for core in range(0, CORES):
            UA_Bdata[core][w] = UAi['B'][core]
            UA_Cdata[core][w] = UAi['C'][core]
        
        UC_Cdata[w] = ts_mcx.C.utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
        
        hdata[w] = ts_mcx.C.max_utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
        Hdata[w] = ts_mcx.C.utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss, heaviest = CORES-1)
        
    
    #----------------------- Create LP Variables -----------------------------#
    problem = LpProblem("Hybrid",LpMinimize)
    
    UA_A = LpVariable.matrix("UA_A", range(CORES), lowBound = 0)
    UA_B = LpVariable.matrix("UA_B", range(CORES), lowBound = 0)
    UA_C = LpVariable.matrix("UA_C", range(CORES), lowBound = 0)
    
    UB_B = LpVariable.matrix("UB_B", range(CORES), lowBound = 0)
    UB_C = LpVariable.matrix("UB_C", range(CORES), lowBound = 0)
    
    UC_C = LpVariable("UC_C", lowBound = 0)
    
    h = LpVariable("h", lowBound = 0)
    H = LpVariable("H", lowBound = 0)
    
    Wa = LpVariable.matrix("Wa", range(CORES), lowBound = 0, cat=vartype)
    Wb = LpVariable.matrix("Wb", range(CORES), lowBound = 0, cat=vartype)
    Wc = LpVariable("Wc", lowBound = 0, cat=vartype);
    
    WO = LpVariable.matrix("WO", range(CORES), lowBound = 0)
    
    Wc.upBound = csch.mW
    
    
    #----------------------- Constraint Set 1 -----------------------------#
    for i, c in enumerate(Wa):
        Wa[i].upBound = csch.mW
        Wb[i].upBound = csch.mW
        WO[i].upBound = csch.mW
        
        if olap:
            problem += Wa[i] + Wc <= csch.mW
            problem += Wb[i] + Wc <= csch.mW
        else:
            problem += Wa[i] + Wb[i] + Wc <= csch.mW
            
        problem += WO[i] >= Wa[i] + Wb[i] + Wc - csch.mW
            
    #----------------------- Constraint Set 2 -----------------------------#
    for i in range(0, CORES):
        UA_A[i].upBound = 1
        UA_B[i].upBound = 1
        UA_C[i].upBound = 1
        
        UB_B[i].upBound = 1
        UB_C[i].upBound = 1
        
        if not len(ts_mcx.A.core(i)) is 0:
            invminp = 1.0/min([t.period for t in ts_mcx.A.core(i)])
        else:
            invminp = 0
            
        tckinfw = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Max']['TCK'])
        tckinfa = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Avg']['TCK'])
        IA_B = ohc.CPMD*Wb[i]*csch.B[i].S*invminp*tckinfw if olap else 0
        IA_C = ohc.CPMDavg*WO[i]*csch.B[i].S*invminp*tckinfa if olap else 0
        
        for j in range(0,csch.mW):
            C1 = (UA_Adata[i][j+1]-UA_Adata[i][j])
            C2 = UA_Adata[i][j]-j*C1
            problem += (UA_A[i]-C1*Wa[i] >= C2)
            C1 = (UA_Bdata[i][j+1]-UA_Bdata[i][j])
            C2 = UA_Bdata[i][j]-j*C1
            problem += (UA_B[i]-C1*Wa[i] >= C2 + IA_B)
            C1 = (UA_Cdata[i][j+1]-UA_Cdata[i][j])
            C2 = UA_Cdata[i][j]-j*C1
            problem += (UA_C[i]-C1*Wa[i] >= C2 + IA_C)
            
            C1 = (UB_Bdata[i][j+1]-UB_Bdata[i][j])
            C2 = UB_Bdata[i][j]-j*C1
            problem += (UB_B[i]-C1*Wb[i] >= C2)
            C1 = (UB_Cdata[i][j+1]-UB_Cdata[i][j])
            C2 = UB_Cdata[i][j]-j*C1
            problem += (UB_C[i]-C1*Wb[i] >= C2)
    
    UC_C.upBound = CORES
    for w in range(0, csch.mW):
        C1 = (UC_Cdata[w+1]-UC_Cdata[w])
        C2 = UC_Cdata[w]-w*C1
        problem += (UC_C-C1*Wc >= C2)
        
    #----------------------- Constraint Set 3 -----------------------------#
        problem += h >= hdata[w] + (Wc-w)*(hdata[w+1]-hdata[w])
        problem += H >= Hdata[w] + (Wc-w)*(Hdata[w+1]-Hdata[w])
        
    #----------------------- Constraint Set 4 -----------------------------#
    
    for c in range(0,CORES):
        problem += UA_A[c] <= 1
        problem += UA_B[c] + UB_B[c] <= 1
        
    problem += lpSum(UA_C) + lpSum(UB_C) + UC_C <= CORES
        
    eps = 1e-6
    
    problem += lpSum(UA_C) + lpSum(UB_C) + (CORES-1)*h + H <= CORES-eps
    
    problem += lpSum(UA_C) + lpSum(UB_C) + UC_C
    
    #----------------------- Solve -----------------------------#
    start = time.clock() 
    status = problem.solve(GUROBI(msg=0, timeLimit = 480))
    fin = time.clock()
    
    rntm = -1
    if LpStatus[status] != 'Not Solved':
        rntm = fin-start

    if LpStatus[status] == 'Optimal':
        ps = scheme([area(Wa[i].varValue, csch.A[i].S) for i in range(0,CORES)],
                    [area(Wb[i].varValue, csch.B[i].S) for i in range(0,CORES)],
                    area(Wc.varValue, MAX_SETS), MAX_WAYS, MAX_SETS)
        
        ps.A = [val.rounded() for val in ps.A]
        ps.B = [val.rounded() for val in ps.B]
        ps.C = ps.C.rounded()
        
        return ps, rntm
    if LpStatus[status] == 'Not Solved':
        return "TimeOut", rntm
        
    return None, rntm



def LP_CBA_OPT(ts, usePST, ab_way_alloc, opt_llc):
    AB = area(ab_way_alloc, 1)
    C = area(16-ab_way_alloc, 16)
    csch = scheme([AB, AB, AB, AB], [AB, AB, AB, AB], C, 16, 16)
    
    
    mdn = len(ts.modes)
    
    capped_wss = {'B' : [(0 if len(ts.level('B').core(c)) == 0 else max([t.wss for t in ts.level('B').core(c)])) for c in range(4)], 
                  'C' : (0 if len(ts.level('C')) == 0 else max([t.wss for t in ts.level('C')]))}
    
    mc_pgs_col = 11
    mc_pgs_os = 40
    ppm = 2048-mc_pgs_col-mc_pgs_os
    problem = LpProblem("Hybrid",LpMinimize)
    
    c0 = 1 if usePST else 0
    
    pp      = LpVariable.dicts("pp",     [(g_i, b_j, c_k) for g_i in range(mdn) for b_j in range(c0, CORES) for c_k in range(16)], lowBound = 0, cat='Continuous')
    use     = LpVariable.dicts("use",    [(g_i, b_j, c_k) for g_i in range(mdn) for b_j in range(c0, CORES) for c_k in range(16)], lowBound = 0, cat='Integer')
    pc      = LpVariable.dicts("pc",     [(g_i, b_j) for g_i in range(mdn) for b_j in range(c0, CORES)], lowBound = 0, cat='Continuous')
    colors  = [[LpAffineExpression() for b_j in range(c0, CORES)] for  g_i in range(mdn)]
    
    colorsSH = LpVariable("colorsSH", lowBound = 1, cat='Integer')
    
    objVar = LpVariable("OBJ", lowBound = 0, cat='Continuous')
    
    #Define "colors"
    for g_i in range(mdn):
        for b_j in range(c0, 4):
            for c_k in range(c0, 16):
                colors[g_i][b_j-c0] += use[g_i, b_j, c_k]

    #Color isolation
    for g_i in range(mdn):
        for c_l in range(16):
            for b_j in range(c0, 3):
                for b_k in range(b_j+1, 4):
                    problem += use[g_i, b_j, c_l] + use[g_i, b_k, c_l] <= 1

    #Isolation between colorsSH and use variables across banks.
    problem += colorsSH >= 1
    if usePST:
        for g_i in range(mdn):
            for c_l in range(16):
                for b_j in range(c0, 4):
                    problem += colorsSH-c_l <= 16*(1-use[g_i, b_j, c_l])
            
    #Express even distribution of pages, using pc.
    for g_i in range(mdn):
        for b_j in range(c0, 4):
            gtsAB = ts.modes[g_i].level('A')
            gtsAB.extend(ts.modes[g_i].level('B'))
            mpt = gtsAB.core(b_j).num_pages()
#             problem += colors[g_i][b_j-c0] >= 1
            for k in range(1, 16):
                problem += pc[g_i, b_j] >= (mpt/k)+(colors[g_i][b_j-c0]-k)*((mpt/(k+1))-(mpt/k))

    #Constrain page count to used colors and even distribution.
    for g_i in range(mdn):
        for b_j in range(c0, 4):
            for c_k in range(16):
                problem += pp[g_i, b_j, c_k] >= pc[g_i, b_j]+(use[g_i, b_j, c_k]-1)*ppm
    
    #DRAM Capacity constraints
    for b_i in range(c0, 4):
        for c_j in range(16):
            p_sum = LpAffineExpression()
            for g_k in range(mdn):
                p_sum += pp[g_k, b_i, c_j]
            problem += p_sum <= ppm
    
    if usePST:
        problem += colorsSH >= ts.sh.num_pages()/ppm
    
#     pp_sum = LpAffineExpression()
#     for g_i in range(mdn):
#         for b_j in range(c0, 4):
#             for c_k in range(16):
#                 pp_sum += pp[g_k, b_i, c_j]
#     problem += pp_sum
    
    ################LLC EXPRESSIONS############Utilization constraints
    if opt_llc:
        UA_Adata = np.zeros((mdn, CORES, csch.mS+1))
        UA_Bdata = np.zeros((mdn, CORES, csch.mS+1))
        UA_Cdata = np.zeros((mdn, CORES, csch.mS+1))
    
        UB_Bdata = np.zeros((mdn, CORES, csch.mS+1))
        UB_Cdata = np.zeros((mdn, CORES, csch.mS+1))
    
        UC_Cdata = np.zeros(mdn)
        hdata = np.zeros(mdn)
        Hdata = np.zeros(mdn)
    
        minAperiods = [[] for i in range(mdn)]
        for g_i in range(mdn):
            for core in range(0, CORES):
                subts = ts.modes[g_i].A.core(core)
                if len(subts) == 0:
                    minAperiods[g_i].append(0)
                else:
                    minAperiods[g_i].append(min([t.period for t in subts]))
    
        for s in range(0,MAX_SETS+1):
            wsch = scheme([area(csch.A[i].W, s) for i in range(0, CORES)],
                          [area(csch.B[i].W, s) for i in range(0, CORES)],
                          csch.C, csch.mW, csch.mS)

            ts.setScheme(wsch)
        
            for g_i in range(mdn):
                for core in range(0, CORES):  
                    gts = ts.modes[g_i]                    
                    UA_Adata[g_i][core][s] = gts.A.core(core).utilization(atlevel = 'A', approximate = True, inflated = True, c_taskset = gts.C, minAperiods = minAperiods[g_i], mx_wss = capped_wss)
                    UA_Bdata[g_i][core][s] = gts.A.core(core).utilization(atlevel = 'B', approximate = True, inflated = True, c_taskset = gts.C, minAperiods = minAperiods[g_i], mx_wss = capped_wss)
                    UA_Cdata[g_i][core][s] = gts.A.core(core).utilization(atlevel = 'C', approximate = True, inflated = True, c_taskset = gts.C, minAperiods = minAperiods[g_i], mx_wss = capped_wss)
            
                    UB_Bdata[g_i][core][s] = gts.B.core(core).utilization(atlevel = 'B', approximate = True, inflated = True, c_taskset = gts.C, minAperiods = minAperiods[g_i], mx_wss = capped_wss)
                    UB_Cdata[g_i][core][s] = gts.B.core(core).utilization(atlevel = 'C', approximate = True, inflated = True, c_taskset = gts.C, minAperiods = minAperiods[g_i], mx_wss = capped_wss)
            
                UAi = defaultdict()
                UAi['B'] = [UA_Bdata[g_i][core][s] for core in range(0, CORES)]
                UAi['C'] = [UA_Cdata[g_i][core][s] for core in range(0, CORES)]
        
                UAi = ts.secondaryInflateConst(UAi, minAperiods = minAperiods[g_i])
                for core in range(0, CORES):
                    UA_Bdata[g_i][core][s] = UAi['B'][core]
                    UA_Cdata[g_i][core][s] = UAi['C'][core]
         
        for g_i in range(mdn):   
            gts = ts.modes[g_i]  
            UC_Cdata[g_i]   = gts.C.utilization     ('C', approximate = True, inflated = True, c_taskset = gts.C, minAperiods = minAperiods, mx_wss = capped_wss)
            hdata[g_i]      = gts.C.max_utilization ('C', approximate = True, inflated = True, c_taskset = gts.C, minAperiods = minAperiods, mx_wss = capped_wss)
            Hdata[g_i]      = gts.C.utilization     ('C', approximate = True, inflated = True, c_taskset = gts.C, minAperiods = minAperiods, mx_wss = capped_wss, heaviest = CORES-1)
    
        UA_A = LpVariable.dicts("UA_A", [(g_i, b_j) for g_i in range(mdn) for b_j in range(CORES)], lowBound = 0, cat='Continuous')
        UA_B = LpVariable.dicts("UA_B", [(g_i, b_j) for g_i in range(mdn) for b_j in range(CORES)], lowBound = 0, cat='Continuous')
        UA_C = LpVariable.dicts("UA_C", [(g_i, b_j) for g_i in range(mdn) for b_j in range(CORES)], lowBound = 0, cat='Continuous')
    
        UB_B = LpVariable.dicts("UB_B", [(g_i, b_j) for g_i in range(mdn) for b_j in range(CORES)], lowBound = 0, cat='Continuous')
        UB_C = LpVariable.dicts("UB_C", [(g_i, b_j) for g_i in range(mdn) for b_j in range(CORES)], lowBound = 0, cat='Continuous')
        
        for g_i in range(mdn):
            for b_j in range(0, 4):
                UA_A[g_i, b_j].upBound = 1
                UA_B[g_i, b_j].upBound = 1
                UA_C[g_i, b_j].upBound = 1
        
                UB_B[g_i, b_j].upBound = 1
                UB_C[g_i, b_j].upBound = 1
            
                if not len(ts.modes[g_i].A.core(i)) is 0:
                    invminp = 1.0/min([t.period for t in ts.modes[g_i].A.core(i)])
                else:
                    invminp = 0
        
                col_var = colorsSH if (usePST and b_j == 0) else colors[g_i][b_j-c0]
            
                tckinfw = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Max']['TCK'])
                tckinfa = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Avg']['TCK'])
                IA_B = ohc.CPMD*csch.B[i].W*col_var*invminp*tckinfw
                IA_C = ohc.CPMDavg*csch.B[i].W*col_var*invminp*tckinfa
                
                for c_k in range(0,MAX_SETS):
                    C1 = (UA_Adata[g_i][b_j][c_k+1]-UA_Adata[g_i][b_j][c_k])
                    C2 = UA_Adata[g_i][b_j][c_k]-c_k*C1
                    problem += (UA_A[g_i, b_j]-C1*col_var >= C2)
                    C1 = (UA_Bdata[g_i][b_j][c_k+1]-UA_Bdata[g_i][b_j][c_k])
                    C2 = UA_Bdata[g_i][b_j][c_k]-c_k*C1+IA_B
                    problem += (UA_B[g_i, b_j]-C1*col_var >= C2)
                    C1 = (UA_Cdata[g_i][b_j][c_k+1]-UA_Cdata[g_i][b_j][c_k])
                    C2 = UA_Cdata[g_i][b_j][c_k]-c_k*C1+IA_C
                    problem += (UA_C[g_i, b_j]-C1*col_var >= C2)
            
                    C1 = (UB_Bdata[g_i][b_j][c_k+1]-UB_Bdata[g_i][b_j][c_k])
                    C2 = UB_Bdata[g_i][b_j][c_k]-c_k*C1
                    problem += (UB_B[g_i, b_j]-C1*col_var >= C2)
                    C1 = (UB_Cdata[g_i][b_j][c_k+1]-UB_Cdata[g_i][b_j][c_k])
                    C2 = UB_Cdata[g_i][b_j][c_k]-c_k*C1
                    problem += (UB_C[g_i, b_j]-C1*col_var >= C2)
                
        #Schedulability Constraints
        for g_i in range(mdn):
            for c in range(0,CORES):
                    problem += UA_A[g_i, c] <= 1
                    problem += UA_B[g_i, c] + UB_B[g_i, c] <= 1
    
        for g_i in range(mdn):
            U_Cm = LpAffineExpression()
            for b_j in range(CORES):
                U_Cm += UA_C[g_i, b_j] + UB_C[g_i, b_j]
        
            eps = 1e-6
            problem += U_Cm + (CORES-1)*hdata[g_i] + Hdata[g_i] <= CORES-eps
        
            U_Cm += UC_Cdata[g_i]        
            problem += U_Cm <= CORES
        
            problem += objVar >= U_Cm
    else:
        if usePST:
            problem += objVar >= colorsSH
            
        for g_i in range(mdn):
            for b_j in range(c0, 4):
                problem += objVar >= colors[g_i][b_j-c0]
    
    problem += objVar
    # print problem 
    start = time.clock() 
#     print "solving"
    status = problem.solve(GUROBI(msg=0, timeLimit = 480))
#     print "solved"
    fin = time.clock()
    
    rntm = -1
    if LpStatus[status] != 'Not Solved':
        rntm = fin-start

    if LpStatus[status] == 'Optimal':
#         print "PAGES PER SLOT:"
#         for b_j in range(c0, 4):
#             print "  CORE " + str(b_j) + ":"
#             print "    pp      :", np.sum([np.sum([use[g_i, b_j, c_k].varValue*pp[g_i, b_j, c_k].varValue for g_i in range(mdn)]) for c_k in range(16)])
#             tsAB = ts.level('A')
#             tsAB.extend(ts.level('B'))
#             mpt = tsAB.core(b_j).num_pages()
#             print "    act_pp  :", mpt
#         if c0 == 1:
#             print " SHARED: " 
#             print "    colorsSH  :", colorsSH.varValue
#         for g_i in range(mdn):
#             print "MODE", g_i
#             print " UTIL:"
#             print "  A:"
#             print "    CORES: " 
#             print "      A:", [UA_A[g_i, b_j].varValue for b_j in range(4)]
#             print "     dA:", [UA_Adata[g_i][b_j][int(colors[g_i][b_j-c0].value())] < UA_A[g_i, b_j].varValue+.0001 for b_j in range(4)]
#             print "      B:", [UA_B[g_i, b_j].varValue for b_j in range(4)]
#             print "     dB:", [UA_Bdata[g_i][b_j][int(colors[g_i][b_j-c0].value())] < UA_B[g_i, b_j].varValue+.0001 for b_j in range(4)]
#             print "      C:", [UA_C[g_i, b_j].varValue for b_j in range(4)]
#             print "     dC:", [UA_Cdata[g_i][b_j][int(colors[g_i][b_j-c0].value())] < UA_C[g_i, b_j].varValue+.0001 for b_j in range(4)]
#             print "    TOTAL: "
#             print "      A:", np.sum([UA_A[g_i, b_j].varValue for b_j in range(4)])
#             print "      B:", np.sum([UA_B[g_i, b_j].varValue for b_j in range(4)])
#             print "      C:", np.sum([UA_C[g_i, b_j].varValue for b_j in range(4)])
#             print "  B:"
#             print "    CORES: " 
#             print "      B:", [UB_B[g_i, b_j].varValue for b_j in range(4)]
#             print "     dB:", [UB_Bdata[g_i][b_j][int(colors[g_i][b_j-c0].value())] < UB_B[g_i, b_j].varValue+.0001 for b_j in range(4)]
#             print "      C:", [UB_C[g_i, b_j].varValue for b_j in range(4)]
#             print "     dC:", [UB_Cdata[g_i][b_j][int(colors[g_i][b_j-c0].value())] < UB_C[g_i, b_j].varValue+.0001 for b_j in range(4)]
#             print "    TOTAL: "
#             print "      B:", np.sum([UB_B[g_i, b_j].varValue for b_j in range(4)])
#             print "      C:", np.sum([UB_C[g_i, b_j].varValue for b_j in range(4)])
#             print "  C:"
#             print "    TOTAL: "
#             print "      C:", UC_Cdata[g_i]
#             print
#             
#             print " PAGES:"
#             print " DISJOINT:"
#             for b_j in range(c0, 4):
#                 print "  CORE " + str(b_j) + ":"
#                 print "    pp      :", [use[g_i, b_j, c_k].varValue*pp[g_i, b_j, c_k].varValue for c_k in range(16)]
#                 print "    use     :", [use[g_i, b_j, c_k].varValue for c_k in range(16)]
#                 print "    pc      :", pc[g_i, b_j].varValue
#                 print "    colors  :", colors[g_i][b_j-c0].value()
#                 gtsAB = ts.modes[g_i].level('A')
#                 gtsAB.extend(ts.modes[g_i].level('B'))
#                 mpt = gtsAB.core(b_j).num_pages()
#                 print "    act_pp  :", mpt
#             if c0 == 1:
#                 print " SHARED: " 
#                 print "    colorsSH  :", colorsSH.varValue
#             print
#         
#         for g_i in range(mdn):
#             print "U_C:", np.sum([UA_C[g_i, b_j].varValue + UB_C[g_i, b_j].varValue for b_j in range(4)])+UC_Cdata[g_i]
#         print "OBJECTIVE: ", objVar.varValue
#         print
#         print


        colorVals = [[int(colors[g_i][b_j-c0].value()) for b_j in range(c0, 4)] for g_i in range(mdn)]
        if usePST:
            colorVals.append(int(colorsSH.varValue))
            
        return colorVals, rntm
            
#         objVar = LpVariable("OBJ", lowBound = 0, cat='Continuous')
#         
#         ps = scheme([area(csch.A[i].W, 0) for i in range(0,CORES)],
#                     [area(csch.B[i].W, 0) for i in range(0,CORES)],
#                     csch.C, MAX_WAYS, MAX_SETS)
#         
#         ps.A = [val.rounded() for val in ps.A]
#         ps.B = [val.rounded() for val in ps.B]
#         ps.C = ps.C.rounded()
#         
#         return ps, rntm
    if LpStatus[status] == 'Not Solved':
        return None, rntm
        
    return None, rntm
    
def LP_WBA2(ts, usePST):
    capped_wss = {'B' : [(0 if len(ts.level('B').core(c)) == 0 else max([t.wss for t in ts.level('B').core(c)])) for c in range(4)], 
                  'C' : (0 if len(ts.level('C')) == 0 else max([t.wss for t in ts.level('C')]))}
                  
    AB = area(0, 16)
    C = area(8, 16)
    csch = scheme([AB, AB, AB, AB], [AB, AB, AB, AB], C, 16, 16)
    
    #----------------------- Gather Data -----------------------------#
    
    UA_Adata = np.zeros((CORES, csch.mW+1))
    UA_Bdata = np.zeros((CORES, csch.mW+1))
    UA_Cdata = np.zeros((CORES, csch.mW+1))
    
    UB_Bdata = np.zeros((CORES, csch.mW+1))
    UB_Cdata = np.zeros((CORES, csch.mW+1))
    
    UC_Cdata = np.zeros((csch.mW+1))
    hdata = np.zeros((csch.mW+1))
    Hdata = np.zeros((csch.mW+1))
    
    minAperiods = []
    for core in range(0, CORES):
        subts = ts.A.core(core)
        if len(subts) == 0:
            minAperiods.append(0)
        else:
            minAperiods.append(min([t.period for t in subts]))
    
    for w in range(0,csch.mW+1):
        wsch = scheme([area(w, csch.A[i].S) for i in range(0, CORES)],
                      [area(w, csch.B[i].S) for i in range(0, CORES)],
                      area(w, csch.C.S), csch.mW, csch.mS)
              
        ts.setScheme(wsch)
        for core in range(0, CORES):              
            UA_Adata[core][w] = ts.A.core(core).utilization(atlevel = 'A', approximate = True, inflated = True, c_taskset = ts.C, minAperiods = minAperiods, mx_wss = capped_wss)
            UA_Bdata[core][w] = ts.A.core(core).utilization(atlevel = 'B', approximate = True, inflated = True, c_taskset = ts.C, minAperiods = minAperiods, mx_wss = capped_wss)
            UA_Cdata[core][w] = ts.A.core(core).utilization(atlevel = 'C', approximate = True, inflated = True, c_taskset = ts.C, minAperiods = minAperiods, mx_wss = capped_wss)
        
            UB_Bdata[core][w] = ts.B.core(core).utilization(atlevel = 'B', approximate = True, inflated = True, c_taskset = ts.C, minAperiods = minAperiods, mx_wss = capped_wss)
            UB_Cdata[core][w] = ts.B.core(core).utilization(atlevel = 'C', approximate = True, inflated = True, c_taskset = ts.C, minAperiods = minAperiods, mx_wss = capped_wss)
        
        UAi = defaultdict()
        UAi['B'] = [UA_Bdata[core][w] for core in range(0, CORES)]
        UAi['C'] = [UA_Cdata[core][w] for core in range(0, CORES)]
    
        UAi = ts.secondaryInflateConst(UAi, minAperiods = minAperiods)
        for core in range(0, CORES):
            UA_Bdata[core][w] = UAi['B'][core]
            UA_Cdata[core][w] = UAi['C'][core]
     
        UC_Cdata[w] = ts.C.utilization('C', approximate = True, inflated = True, c_taskset = ts.C, minAperiods = minAperiods, mx_wss = capped_wss)
    
        hdata[w] = ts.C.max_utilization('C', approximate = True, inflated = True, c_taskset = ts.C, minAperiods = minAperiods, mx_wss = capped_wss)
        Hdata[w] = ts.C.utilization('C', approximate = True, inflated = True, c_taskset = ts.C, minAperiods = minAperiods, mx_wss = capped_wss, heaviest = CORES-1)
     
     
    #----------------------- Create LP Variables -----------------------------#
    problem = LpProblem("Hybrid",LpMinimize)
    
    UA_A = LpVariable.matrix("UA_A", range(CORES), lowBound = 0)
    UA_B = LpVariable.matrix("UA_B", range(CORES), lowBound = 0)
    UA_C = LpVariable.matrix("UA_C", range(CORES), lowBound = 0)
    
    UB_B = LpVariable.matrix("UB_B", range(CORES), lowBound = 0)
    UB_C = LpVariable.matrix("UB_C", range(CORES), lowBound = 0)
        
    UC_C = LpVariable("UC_C", lowBound = 0)
    
    h = LpVariable("h", lowBound = 0)
    H = LpVariable("H", lowBound = 0)
    
    Wab = LpVariable.matrix("Wab", range(CORES), lowBound = 0, cat='Integer')
    Wc = LpVariable("Wc", lowBound = 0, cat = 'Integer')
    
    #----------------------- Constraint Set 1 -----------------------------#
    Wc.upBound = csch.mW
    for b_j in range(4):
        Wab[b_j].upBound = csch.mW
    problem += lpSum(Wab) + Wc <= csch.mW
            
    
    #----------------------- Constraint Set 2 -----------------------------#
    for b_j in range(0, CORES):
        UA_A[b_j].upBound = 1
        UA_B[b_j].upBound = 1
        UA_C[b_j].upBound = 1
    
        UB_B[b_j].upBound = 1
        UB_C[b_j].upBound = 1
        
        if not len(ts.A.core(i)) is 0:
            invminp = 1.0/min([t.period for t in ts.A.core(i)])
        else:
            invminp = 0
            
        tckinfw = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Max']['TCK'])
        tckinfa = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Avg']['TCK'])
        
        w_var = Wab[b_j]
        
        IA_B = ohc.CPMD*w_var*csch.B[i].S*invminp*tckinfw
        IA_C = ohc.CPMDavg*w_var*csch.B[i].S*invminp*tckinfa
        
        
        for w_k in range(0,csch.mW):
            C1 = (UA_Adata[b_j][w_k+1]-UA_Adata[b_j][w_k])
            C2 = UA_Adata[b_j][w_k]-w_k*C1
            problem += (UA_A[b_j]-C1*w_var >= C2)
            C1 = (UA_Bdata[b_j][w_k+1]-UA_Bdata[b_j][w_k])
            C2 = UA_Bdata[b_j][w_k]-w_k*C1
            problem += (UA_B[b_j]-C1*w_var >= C2 + IA_B)
            C1 = (UA_Cdata[b_j][w_k+1]-UA_Cdata[b_j][w_k])
            C2 = UA_Cdata[b_j][w_k]-w_k*C1
            problem += (UA_C[b_j]-C1*w_var >= C2 + IA_C)
        
            C1 = (UB_Bdata[b_j][w_k+1]-UB_Bdata[b_j][w_k])
            C2 = UB_Bdata[b_j][w_k]-w_k*C1
            problem += (UB_B[b_j]-C1*w_var >= C2)
            C1 = (UB_Cdata[b_j][w_k+1]-UB_Cdata[b_j][w_k])
            C2 = UB_Cdata[b_j][w_k]-w_k*C1
            problem += (UB_C[b_j]-C1*w_var >= C2)

    UC_C.upBound = CORES
    for w_k in range(0, csch.mW):
        C1 = (UC_Cdata[w_k+1]-UC_Cdata[w_k])
        C2 = UC_Cdata[w_k]-w_k*C1
        problem += (UC_C-C1*Wc >= C2)
    
    #----------------------- Constraint Set 3 -----------------------------#
        problem += h >= hdata[w_k] + (Wc-w_k)*(hdata[w_k+1]-hdata[w_k])
        problem += H >= Hdata[w_k] + (Wc-w_k)*(Hdata[w_k+1]-Hdata[w_k])
        
    #----------------------- Constraint Set 4 -----------------------------#
    
    for c in range(0,CORES):
        problem += UA_A[c] <= 1
        problem += UA_B[c] + UB_B[c] <= 1
        
    problem += lpSum(UA_C) + lpSum(UB_C) + UC_C <= CORES
        
    eps = 1e-6
    
    problem += lpSum(UA_C) + lpSum(UB_C) + (CORES-1)*h + H <= CORES-eps
    
    problem += lpSum(UA_C) + lpSum(UB_C) + UC_C
    
    
    #----------------------- Solve -----------------------------#
    start = time.clock() 
    status = problem.solve(GUROBI(msg=0, timeLimit = 480))
    fin = time.clock()
    
    rntm = -1
    if LpStatus[status] != 'Not Solved':
        rntm = fin-start

    if LpStatus[status] == 'Optimal':
#         for g_i in range(mdn):
#             WabV = []
#             if usePST:
#                 WabV.append(int(Wsh.varValue))
#             WabV.extend([int(Wab[g_i, b_j].varValue) for b_j in range(c0, 4)])
#             print "ALLOC"
#             print "  Wh:", WabV
#             print "  Wc:", Wc[g_i].varValue
#             print "MODE", g_i
#             print " UTIL:"
#             print "  A:"
#             print "    CORES: " 
#             print "      A:", [UA_A[g_i, b_j].varValue for b_j in range(4)]
#             print "     dA:", [UA_Adata[g_i][b_j][WabV[b_j]] < UA_A[g_i, b_j].varValue+.0001 for b_j in range(4)]
#             print "      B:", [UA_B[g_i, b_j].varValue for b_j in range(4)]
#             print "     dB:", [UA_Bdata[g_i][b_j][WabV[b_j]] < UA_B[g_i, b_j].varValue+.0001 for b_j in range(4)]
#             print "      C:", [UA_C[g_i, b_j].varValue for b_j in range(4)]
#             print "     dC:", [UA_Cdata[g_i][b_j][WabV[b_j]] < UA_C[g_i, b_j].varValue+.0001 for b_j in range(4)]
#             print "    TOTAL: "
#             print "      A:", np.sum([UA_A[g_i, b_j].varValue for b_j in range(4)])
#             print "      B:", np.sum([UA_B[g_i, b_j].varValue for b_j in range(4)])
#             print "      C:", np.sum([UA_C[g_i, b_j].varValue for b_j in range(4)])
#             print "  B:"
#             print "    CORES: " 
#             print "      B:", [UB_B[g_i, b_j].varValue for b_j in range(4)]
#             print "     dB:", [UB_Bdata[g_i][b_j][WabV[b_j]] < UB_B[g_i, b_j].varValue+.0001 for b_j in range(4)]
#             print "      C:", [UB_C[g_i, b_j].varValue for b_j in range(4)]
#             print "     dC:", [UB_Cdata[g_i][b_j][WabV[b_j]] < UB_C[g_i, b_j].varValue+.0001 for b_j in range(4)]
#             print "    TOTAL: "
#             print "      B:", np.sum([UB_B[g_i, b_j].varValue for b_j in range(4)])
#             print "      C:", np.sum([UB_C[g_i, b_j].varValue for b_j in range(4)])
#             print "  C:"
#             print "    TOTAL: "
#             print "      C:", UC_C[g_i].varValue
#             print "     dC:", UC_Cdata[g_i][int(Wc[g_i].varValue)] < UC_C[g_i].varValue+.0001
#             print
#         
#         for g_i in range(mdn):
#             print "U_C:", np.sum([UA_C[g_i, b_j].varValue + UB_C[g_i, b_j].varValue for b_j in range(4)])+UC_C[g_i].varValue
#         print "OBJECTIVE: ", objVar.varValue
#         print
#         print

        ps = scheme([area(0, csch.A[i].S) for i in range(0,CORES)],
                    [area(0, csch.B[i].S) for i in range(0,CORES)],
                    area(0, MAX_SETS), MAX_WAYS, MAX_SETS)
        
        ps.A = [val.rounded() for val in ps.A]
        ps.B = [val.rounded() for val in ps.B]
        ps.C = ps.C.rounded()
        
        return ps, rntm
    if LpStatus[status] == 'Not Solved':
        return None, rntm
        
    return None, rntm
    
# def LP_WBA(ts, usePST):
#     capped_wss = {'B' : [(0 if len(ts.level('B').core(c)) == 0 else max([t.wss for t in ts.level('B').core(c)])) for c in range(4)], 
#                   'C' : (0 if len(ts.level('C')) == 0 else max([t.wss for t in ts.level('C')]))}
#     AB = area(0, 16)
#     C = area(8, 16)
#     csch = scheme([AB, AB, AB, AB], [AB, AB, AB, AB], C, 16, 16)
#     
#     mdn = len(ts.modes)
#     
#     #----------------------- Gather Data -----------------------------#
#     
#     UA_Adata = np.zeros((mdn, CORES, csch.mW+1))
#     UA_Bdata = np.zeros((mdn, CORES, csch.mW+1))
#     UA_Cdata = np.zeros((mdn, CORES, csch.mW+1))
#     
#     UB_Bdata = np.zeros((mdn, CORES, csch.mW+1))
#     UB_Cdata = np.zeros((mdn, CORES, csch.mW+1))
#     
#     UC_Cdata = np.zeros((mdn, csch.mW+1))
#     hdata = np.zeros((mdn, csch.mW+1))
#     Hdata = np.zeros((mdn, csch.mW+1))
#     
#     minAperiods = [[] for i in range(mdn)]
#     for g_i in range(mdn):
#         for core in range(0, CORES):
#             subts = ts.modes[g_i].A.core(core)
#             if len(subts) == 0:
#                 minAperiods[g_i].append(0)
#             else:
#                 minAperiods[g_i].append(min([t.period for t in subts]))
#     
#     for w in range(0,csch.mW+1):
#         wsch = scheme([area(w, csch.A[i].S) for i in range(0, CORES)],
#                       [area(w, csch.B[i].S) for i in range(0, CORES)],
#                       area(w, csch.C.S), csch.mW, csch.mS)
#               
#         ts.setScheme(wsch)
#         for g_i in range(mdn):
#             gts = ts.modes[g_i]    
#             for core in range(0, CORES):              
#                 UA_Adata[g_i][core][w] = gts.A.core(core).utilization(atlevel = 'A', approximate = True, inflated = True, c_taskset = gts.C, minAperiods = minAperiods[g_i], mx_wss = capped_wss)
#                 UA_Bdata[g_i][core][w] = gts.A.core(core).utilization(atlevel = 'B', approximate = True, inflated = True, c_taskset = gts.C, minAperiods = minAperiods[g_i], mx_wss = capped_wss)
#                 UA_Cdata[g_i][core][w] = gts.A.core(core).utilization(atlevel = 'C', approximate = True, inflated = True, c_taskset = gts.C, minAperiods = minAperiods[g_i], mx_wss = capped_wss)
#             
#                 UB_Bdata[g_i][core][w] = gts.B.core(core).utilization(atlevel = 'B', approximate = True, inflated = True, c_taskset = gts.C, minAperiods = minAperiods[g_i], mx_wss = capped_wss)
#                 UB_Cdata[g_i][core][w] = gts.B.core(core).utilization(atlevel = 'C', approximate = True, inflated = True, c_taskset = gts.C, minAperiods = minAperiods[g_i], mx_wss = capped_wss)
#             
#             UAi = defaultdict()
#             UAi['B'] = [UA_Bdata[g_i][core][w] for core in range(0, CORES)]
#             UAi['C'] = [UA_Cdata[g_i][core][w] for core in range(0, CORES)]
#         
#             UAi = ts.secondaryInflateConst(UAi, minAperiods = minAperiods[g_i])
#             for core in range(0, CORES):
#                 UA_Bdata[g_i][core][w] = UAi['B'][core]
#                 UA_Cdata[g_i][core][w] = UAi['C'][core]
#          
#             UC_Cdata[g_i, w] = gts.C.utilization('C', approximate = True, inflated = True, c_taskset = gts.C, minAperiods = minAperiods, mx_wss = capped_wss)
#         
#             hdata[g_i, w] = gts.C.max_utilization('C', approximate = True, inflated = True, c_taskset = gts.C, minAperiods = minAperiods, mx_wss = capped_wss)
#             Hdata[g_i, w] = gts.C.utilization('C', approximate = True, inflated = True, c_taskset = gts.C, minAperiods = minAperiods, mx_wss = capped_wss, heaviest = CORES-1)
#      
#      
#     #----------------------- Create LP Variables -----------------------------#
#     problem = LpProblem("Hybrid",LpMinimize)
#     
#     UA_A = LpVariable.dicts("UA_A", [(g_i, b_j) for g_i in range(mdn) for b_j in range(CORES)], lowBound = 0, cat='Continuous')
#     UA_B = LpVariable.dicts("UA_B", [(g_i, b_j) for g_i in range(mdn) for b_j in range(CORES)], lowBound = 0, cat='Continuous')
#     UA_C = LpVariable.dicts("UA_C", [(g_i, b_j) for g_i in range(mdn) for b_j in range(CORES)], lowBound = 0, cat='Continuous')
#     
#     UB_B = LpVariable.dicts("UB_B", [(g_i, b_j) for g_i in range(mdn) for b_j in range(CORES)], lowBound = 0, cat='Continuous')
#     UB_C = LpVariable.dicts("UB_C", [(g_i, b_j) for g_i in range(mdn) for b_j in range(CORES)], lowBound = 0, cat='Continuous')
#         
#     UC_C = LpVariable.matrix("UC_C", range(mdn), lowBound = 0, cat='Continuous')
#     
#     h = LpVariable.matrix("h", range(mdn), lowBound = 0, cat='Continuous')
#     H = LpVariable.matrix("H", range(mdn), lowBound = 0, cat='Continuous')
#     
#     objVar = LpVariable("OBJ", lowBound = 0, cat='Continuous')
#     
#     c0 = 1 if usePST else 0
#     
#     Wsh = LpVariable("Wsh", lowBound = 0, cat='Integer')
#     Wab = LpVariable.dicts("Wab", [(g_i, b_j) for g_i in range(mdn) for b_j in range(c0, CORES)], lowBound = 0, cat='Integer')
#     Wc = LpVariable.matrix("Wc", range(mdn), lowBound = 0, cat='Integer')
#     
#     Wsh.upBound = csch.mW
#     
#     #----------------------- Constraint Set 1 -----------------------------#
#     for g_i in range(mdn):
#         Wc[g_i].upBound = csch.mW
#         way_sum = LpAffineExpression()
#         for b_j in range(c0,4):
#             Wab[g_i, b_j].upBound = csch.mW
#             way_sum += Wab[g_i, b_j]
#         if usePST:
#             problem += Wsh + way_sum + Wc[g_i] <= csch.mW
#         else:
#             problem += way_sum + Wc[g_i] <= csch.mW
#             
#     
#     #----------------------- Constraint Set 2 -----------------------------#
#     for g_i in range(mdn):
#         tsg = ts.modes[g_i]
#         for b_j in range(0, CORES):
#             UA_A[g_i, b_j].upBound = 1
#             UA_B[g_i, b_j].upBound = 1
#             UA_C[g_i, b_j].upBound = 1
#         
#             UB_B[g_i, b_j].upBound = 1
#             UB_C[g_i, b_j].upBound = 1
#             
#             if not len(gts.A.core(i)) is 0:
#                 invminp = 1.0/min([t.period for t in gts.A.core(i)])
#             else:
#                 invminp = 0
#                 
#             tckinfw = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Max']['TCK'])
#             tckinfa = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Avg']['TCK'])
#             
#             w_var = Wsh 
#             if not (usePST and b_j == 0):
#                 w_var = Wab[g_i, b_j]
#             
#             IA_B = ohc.CPMD*w_var*csch.B[i].S*invminp*tckinfw
#             IA_C = ohc.CPMDavg*w_var*csch.B[i].S*invminp*tckinfa
#             
#             
#             for w_k in range(0,csch.mW):
#                 C1 = (UA_Adata[g_i][b_j][w_k+1]-UA_Adata[g_i][b_j][w_k])
#                 C2 = UA_Adata[g_i][b_j][w_k]-w_k*C1
#                 problem += (UA_A[g_i, b_j]-C1*w_var >= C2)
#                 C1 = (UA_Bdata[g_i][b_j][w_k+1]-UA_Bdata[g_i][b_j][w_k])
#                 C2 = UA_Bdata[g_i][b_j][w_k]-w_k*C1
#                 problem += (UA_B[g_i, b_j]-C1*w_var >= C2 + IA_B)
#                 C1 = (UA_Cdata[g_i][b_j][w_k+1]-UA_Cdata[g_i][b_j][w_k])
#                 C2 = UA_Cdata[g_i][b_j][w_k]-w_k*C1
#                 problem += (UA_C[g_i, b_j]-C1*w_var >= C2 + IA_C)
#             
#                 C1 = (UB_Bdata[g_i][b_j][w_k+1]-UB_Bdata[g_i][b_j][w_k])
#                 C2 = UB_Bdata[g_i][b_j][w_k]-w_k*C1
#                 problem += (UB_B[g_i, b_j]-C1*w_var >= C2)
#                 C1 = (UB_Cdata[g_i][b_j][w_k+1]-UB_Cdata[g_i][b_j][w_k])
#                 C2 = UB_Cdata[g_i][b_j][w_k]-w_k*C1
#                 problem += (UB_C[g_i, b_j]-C1*w_var >= C2)
#     
#         UC_C[g_i].upBound = CORES
#         for w_k in range(0, csch.mW):
#             C1 = (UC_Cdata[g_i][w_k+1]-UC_Cdata[g_i][w_k])
#             C2 = UC_Cdata[g_i][w_k]-w_k*C1
#             problem += (UC_C[g_i]-C1*Wc[g_i] >= C2)
#     
#     #----------------------- Constraint Set 3 -----------------------------#
#             problem += h[g_i] >= hdata[g_i][w_k] + (Wc[g_i]-w_k)*(hdata[g_i][w_k+1]-hdata[g_i][w_k])
#             problem += H[g_i] >= Hdata[g_i][w_k] + (Wc[g_i]-w_k)*(Hdata[g_i][w_k+1]-Hdata[g_i][w_k])
#         
#     #----------------------- Constraint Set 4 -----------------------------#
#         gUAB_C = LpAffineExpression()
#         for b_j in range(0,CORES):
#             problem += UA_A[g_i, b_j] <= 1
#             problem += UA_B[g_i, b_j] + UB_B[g_i, b_j] <= 1
#             
#             gUAB_C += UA_C[g_i, b_j] + UB_C[g_i, b_j]
#         
#         problem += gUAB_C + UC_C[g_i] <= CORES
#         
#         eps = 1e-6
#     
#         problem += gUAB_C + (CORES-1)*h[g_i] + H[g_i] <= CORES-eps
#         objVar >= gUAB_C + UC_C[g_i]
#         
#     problem += objVar
#     
#     #----------------------- Solve -----------------------------#
#     start = time.clock() 
#     status = problem.solve(GUROBI(msg=0, timeLimit = 480))
#     fin = time.clock()
#     
#     rntm = -1
#     if LpStatus[status] != 'Not Solved':
#         rntm = fin-start
# 
#     if LpStatus[status] == 'Optimal':
# #         for g_i in range(mdn):
# #             WabV = []
# #             if usePST:
# #                 WabV.append(int(Wsh.varValue))
# #             WabV.extend([int(Wab[g_i, b_j].varValue) for b_j in range(c0, 4)])
# #             print "ALLOC"
# #             print "  Wh:", WabV
# #             print "  Wc:", Wc[g_i].varValue
# #             print "MODE", g_i
# #             print " UTIL:"
# #             print "  A:"
# #             print "    CORES: " 
# #             print "      A:", [UA_A[g_i, b_j].varValue for b_j in range(4)]
# #             print "     dA:", [UA_Adata[g_i][b_j][WabV[b_j]] < UA_A[g_i, b_j].varValue+.0001 for b_j in range(4)]
# #             print "      B:", [UA_B[g_i, b_j].varValue for b_j in range(4)]
# #             print "     dB:", [UA_Bdata[g_i][b_j][WabV[b_j]] < UA_B[g_i, b_j].varValue+.0001 for b_j in range(4)]
# #             print "      C:", [UA_C[g_i, b_j].varValue for b_j in range(4)]
# #             print "     dC:", [UA_Cdata[g_i][b_j][WabV[b_j]] < UA_C[g_i, b_j].varValue+.0001 for b_j in range(4)]
# #             print "    TOTAL: "
# #             print "      A:", np.sum([UA_A[g_i, b_j].varValue for b_j in range(4)])
# #             print "      B:", np.sum([UA_B[g_i, b_j].varValue for b_j in range(4)])
# #             print "      C:", np.sum([UA_C[g_i, b_j].varValue for b_j in range(4)])
# #             print "  B:"
# #             print "    CORES: " 
# #             print "      B:", [UB_B[g_i, b_j].varValue for b_j in range(4)]
# #             print "     dB:", [UB_Bdata[g_i][b_j][WabV[b_j]] < UB_B[g_i, b_j].varValue+.0001 for b_j in range(4)]
# #             print "      C:", [UB_C[g_i, b_j].varValue for b_j in range(4)]
# #             print "     dC:", [UB_Cdata[g_i][b_j][WabV[b_j]] < UB_C[g_i, b_j].varValue+.0001 for b_j in range(4)]
# #             print "    TOTAL: "
# #             print "      B:", np.sum([UB_B[g_i, b_j].varValue for b_j in range(4)])
# #             print "      C:", np.sum([UB_C[g_i, b_j].varValue for b_j in range(4)])
# #             print "  C:"
# #             print "    TOTAL: "
# #             print "      C:", UC_C[g_i].varValue
# #             print "     dC:", UC_Cdata[g_i][int(Wc[g_i].varValue)] < UC_C[g_i].varValue+.0001
# #             print
# #         
# #         for g_i in range(mdn):
# #             print "U_C:", np.sum([UA_C[g_i, b_j].varValue + UB_C[g_i, b_j].varValue for b_j in range(4)])+UC_C[g_i].varValue
# #         print "OBJECTIVE: ", objVar.varValue
# #         print
# #         print
# 
#         ps = scheme([area(0, csch.A[i].S) for i in range(0,CORES)],
#                     [area(0, csch.B[i].S) for i in range(0,CORES)],
#                     area(0, MAX_SETS), MAX_WAYS, MAX_SETS)
#         
#         ps.A = [val.rounded() for val in ps.A]
#         ps.B = [val.rounded() for val in ps.B]
#         ps.C = ps.C.rounded()
#         
#         return ps, rntm
#     if LpStatus[status] == 'Not Solved':
#         return None, rntm
#         
#     return None, rntm
    
    
# def LP_CBA_Hybrid(ts, usePST, ab_way_alloc):
#     #mode change panges per color on a bank
#     AB = area(ab_way_alloc, 1)
#     C = area(16-ab_way_alloc, 16)
#     csch = scheme([AB, AB, AB, AB], [AB, AB, AB, AB], C, 16, 16)
#     
#     mdn = len(ts.modes)
#     
#     sched = ts.get_sched_all()
#     
#     UA_Adata = np.zeros((mdn, CORES, csch.mS+1))
#     UA_Bdata = np.zeros((mdn, CORES, csch.mS+1))
#     UA_Cdata = np.zeros((mdn, CORES, csch.mS+1))
#     
#     UB_Bdata = np.zeros((mdn, CORES, csch.mS+1))
#     UB_Cdata = np.zeros((mdn, CORES, csch.mS+1))
#     
#     UC_Cdata = np.zeros(mdn)
#     hdata = np.zeros(mdn)
#     Hdata = np.zeros(mdn)
#     
#     minAperiods = [[] for i in range(mdn)]
#     for g_i in range(mdn):
#         for core in range(0, CORES):
#             subts = ts.modes[g_i].A.core(core)
#             if len(subts) == 0:
#                 minAperiods[g_i].append(0)
#             else:
#                 minAperiods[g_i].append(min([t.period for t in subts]))
#     
#     for s in range(0,MAX_SETS+1):
#     	wsch = scheme([area(csch.A[i].W, s) for i in range(0, CORES)],
#                       [area(csch.B[i].W, s) for i in range(0, CORES)],
#                       csch.C, csch.mW, csch.mS)
# 
#         ts.setScheme(wsch)
#         
#         for g_i in range(mdn):
#             for core in range(0, CORES):  
#                 gts = ts.modes[g_i]                    
#                 UA_Adata[g_i][core][s] = gts.A.core(core).utilization(atlevel = 'A', approximate = True, inflated = True, c_taskset = gts.C, minAperiods = minAperiods[g_i], mx_wss = capped_wss)
#                 UA_Bdata[g_i][core][s] = gts.A.core(core).utilization(atlevel = 'B', approximate = True, inflated = True, c_taskset = gts.C, minAperiods = minAperiods[g_i], mx_wss = capped_wss)
#                 UA_Cdata[g_i][core][s] = gts.A.core(core).utilization(atlevel = 'C', approximate = True, inflated = True, c_taskset = gts.C, minAperiods = minAperiods[g_i], mx_wss = capped_wss)
#             
#                 UB_Bdata[g_i][core][s] = gts.B.core(core).utilization(atlevel = 'B', approximate = True, inflated = True, c_taskset = gts.C, minAperiods = minAperiods[g_i], mx_wss = capped_wss)
#                 UB_Cdata[g_i][core][s] = gts.B.core(core).utilization(atlevel = 'C', approximate = True, inflated = True, c_taskset = gts.C, minAperiods = minAperiods[g_i], mx_wss = capped_wss)
#             
#             UAi = defaultdict()
#             UAi['B'] = [UA_Bdata[g_i][core][s] for core in range(0, CORES)]
#             UAi['C'] = [UA_Cdata[g_i][core][s] for core in range(0, CORES)]
#         
#             UAi = ts.secondaryInflateConst(UAi, minAperiods = minAperiods[g_i])
#             for core in range(0, CORES):
#                 UA_Bdata[g_i][core][s] = UAi['B'][core]
#                 UA_Cdata[g_i][core][s] = UAi['C'][core]
#          
#     for g_i in range(mdn):   
#         gts = ts.modes[g_i]  
#         UC_Cdata[g_i]   = gts.C.utilization     ('C', approximate = True, inflated = True, c_taskset = gts.C, minAperiods = minAperiods, mx_wss = capped_wss)
#         hdata[g_i]      = gts.C.max_utilization ('C', approximate = True, inflated = True, c_taskset = gts.C, minAperiods = minAperiods, mx_wss = capped_wss)
#         Hdata[g_i]      = gts.C.utilization     ('C', approximate = True, inflated = True, c_taskset = gts.C, minAperiods = minAperiods, mx_wss = capped_wss, heaviest = CORES-1)
#     
#     mc_pgs_col = 11
#     mc_pgs_os = 40
#     ppm = 2048-mc_pgs_col-mc_pgs_os
#     problem = LpProblem("Hybrid",LpMinimize)
#     
#     UA_A = LpVariable.dicts("UA_A", [(g_i, b_j) for g_i in range(mdn) for b_j in range(CORES)], lowBound = 0, cat='Continuous')
#     UA_B = LpVariable.dicts("UA_B", [(g_i, b_j) for g_i in range(mdn) for b_j in range(CORES)], lowBound = 0, cat='Continuous')
#     UA_C = LpVariable.dicts("UA_C", [(g_i, b_j) for g_i in range(mdn) for b_j in range(CORES)], lowBound = 0, cat='Continuous')
#     
#     UB_B = LpVariable.dicts("UB_B", [(g_i, b_j) for g_i in range(mdn) for b_j in range(CORES)], lowBound = 0, cat='Continuous')
#     UB_C = LpVariable.dicts("UB_C", [(g_i, b_j) for g_i in range(mdn) for b_j in range(CORES)], lowBound = 0, cat='Continuous')
#     
#     c0 = 1 if usePST else 0
#     
#     pp      = LpVariable.dicts("pp",     [(g_i, b_j, c_k) for g_i in range(mdn) for b_j in range(c0, CORES) for c_k in range(16)], lowBound = 0, cat='Continuous')
#     pc      = LpVariable.dicts("pc",     [(g_i, b_j) for g_i in range(mdn) for b_j in range(c0, CORES)], lowBound = 0, cat='Continuous')
#     colors  = LpVariable.dicts("colors", [(g_i, b_j) for g_i in range(mdn) for b_j in range(c0, CORES)], lowBound = 0, cat='Continuous')
#     
#     colorsSH = LpVariable("colorsSH", lowBound = 1, cat='Integer')
#     
#     objVar = LpVariable("OBJ", lowBound = 0, cat='Continuous')
# 
#     #Color isolation
#     for g_i in range(mdn):
#         col_sum = LpAffineExpression()
#         if usePST:
#             col_sum += colorSH
#         for b_j in range(c0, 4):
#             col_sum += colors[g_i, b_j]
#         problem += col_sum <= 16
# 
#     problem += colorsSH >= 1
#             
#     #Express even distribution of pages, using pc.
#     for g_i in range(mdn):
#         for b_j in range(c0, 4):
#             gtsAB = ts.modes[g_i].level('A')
#             gtsAB.extend(ts.modes[g_i].level('B'))
#             mpt = gtsAB.core(b_j).num_pages()
# #             problem += colors[g_i][b_j-c0] >= 1
#             for k in range(1, 16):
#                 problem += pc[g_i, b_j] >= (mpt/k)+(colors[g_i][b_j-c0]-k)*((mpt/(k+1))-(mpt/k))
# 
# 
#     #two modes, F and G
#     #Apply a shift to where sums start and stop
#     #4, 1 
#     #F.use(b_i, c_j) = Sum_(k < i)(F.colors(b_k)) <= c_j < Sum_(k < i)(F.colors(b_k)) + F.colors(b_i)
#     #F.nuse(b_i, c_j) >= Sum_(k < i)(F.colors(b_k)) - c_j
#     #F.nuse(b_i, c_j) >= -Sum_(k <= i)(F.colors(b_k)) + c_j
#     #shift s
#     #F.nuse(b_i, c_j) >= Sum_(k > s & k < i)(F.colors(b_k)) - c_j
#     #problem += pp[g_i, b_j, c_k] >= pc[g_i, b_j]-nuse[g_i, b_j, c_k]*ppm
#     #k % 4
#     #k >= 0, k <= i
#     #k <= i-s
#     #i = 3, s = 2
#     
#     for g_i in range(mdn):
#         for b_j in range(c0, 4):
#             nuse(b_i, c_j) >= Sum_(k < i)(F.colors(b_k)) - c_j
#                 
#     for g_i in range(mdn):
#         for b_j in range(c0, 4):
#             kend = b_j >= s ? b_j : b_j+4
#             bsum = LpAffineExpression()
#             if b_j < s:
#                 for k in range(s, kend):
#                     bsum += colors[g_i][k % 4]
#             for c_k in range(16):
#                 nuse >= bsum - c_j
#                 nuse >= c_j - bsum + colors[g_i][b_j] - 1
#                 problem += pp[g_i, b_j, c_k] >= pc[g_i, b_j]-nuse[g_i, b_j, c_k]*ppm
#     
#     #Constrain page count to used colors and even distribution.
#     for g_i in range(mdn):
#         for b_j in range(c0, 4):
#             for c_k in range(16):
#                 problem += pp[g_i, b_j, c_k] >= pc[g_i, b_j]+(use[g_i, b_j, c_k]-1)*ppm
#     
#     #DRAM Capacity constraints
#     for b_i in range(c0, 4):
#         for c_j in range(16):
#             p_sum = LpAffineExpression()
#             for g_k in range(mdn):
#                 p_sum += pp[g_k, b_i, c_j]
#             problem += p_sum <= ppm
#     
#     if usePST:
#         problem += colorsSH >= ts.sh.num_pages()/ppm
#     
# #     pp_sum = LpAffineExpression()
# #     for g_i in range(mdn):
# #         for b_j in range(c0, 4):
# #             for c_k in range(16):
# #                 pp_sum += pp[g_k, b_i, c_j]
# #     problem += pp_sum
#     
#     #Utilization constraints
#     for g_i in range(mdn):
#         for b_j in range(0, 4):
#             UA_A[g_i, b_j].upBound = 1
#             UA_B[g_i, b_j].upBound = 1
#             UA_C[g_i, b_j].upBound = 1
#         
#             UB_B[g_i, b_j].upBound = 1
#             UB_C[g_i, b_j].upBound = 1
#             
#             if not len(ts.modes[g_i].A.core(i)) is 0:
#                 invminp = 1.0/min([t.period for t in ts.modes[g_i].A.core(i)])
#             else:
#                 invminp = 0
#         
#             col_var = colorsSH if (usePST and b_j == 0) else colors[g_i][b_j-c0]
#             
#             tckinfw = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Max']['TCK'])
#             tckinfa = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Avg']['TCK'])
#             IA_B = ohc.CPMD*csch.B[i].W*col_var*invminp*tckinfw
#             IA_C = ohc.CPMDavg*csch.B[i].W*col_var*invminp*tckinfa
#                 
#             for c_k in range(0,MAX_SETS):
#                 C1 = (UA_Adata[g_i][b_j][c_k+1]-UA_Adata[g_i][b_j][c_k])
#                 C2 = UA_Adata[g_i][b_j][c_k]-c_k*C1
#                 problem += (UA_A[g_i, b_j]-C1*col_var >= C2)
#                 C1 = (UA_Bdata[g_i][b_j][c_k+1]-UA_Bdata[g_i][b_j][c_k])
#                 C2 = UA_Bdata[g_i][b_j][c_k]-c_k*C1+IA_B
#                 problem += (UA_B[g_i, b_j]-C1*col_var >= C2)
#                 C1 = (UA_Cdata[g_i][b_j][c_k+1]-UA_Cdata[g_i][b_j][c_k])
#                 C2 = UA_Cdata[g_i][b_j][c_k]-c_k*C1+IA_C
#                 problem += (UA_C[g_i, b_j]-C1*col_var >= C2)
#             
#                 C1 = (UB_Bdata[g_i][b_j][c_k+1]-UB_Bdata[g_i][b_j][c_k])
#                 C2 = UB_Bdata[g_i][b_j][c_k]-c_k*C1
#                 problem += (UB_B[g_i, b_j]-C1*col_var >= C2)
#                 C1 = (UB_Cdata[g_i][b_j][c_k+1]-UB_Cdata[g_i][b_j][c_k])
#                 C2 = UB_Cdata[g_i][b_j][c_k]-c_k*C1
#                 problem += (UB_C[g_i, b_j]-C1*col_var >= C2)
#                 
#     #Schedulability Constraints
#     for g_i in range(mdn):
#         for c in range(0,CORES):
#                 problem += UA_A[g_i, c] <= 1
#                 problem += UA_B[g_i, c] + UB_B[g_i, c] <= 1
#     
#     for g_i in range(mdn):
#         U_Cm = LpAffineExpression()
#         for b_j in range(CORES):
#             U_Cm += UA_C[g_i, b_j] + UB_C[g_i, b_j]
#         
#         eps = 1e-6
#         problem += U_Cm + (CORES-1)*hdata[g_i] + Hdata[g_i] <= CORES-eps
#         
#         U_Cm += UC_Cdata[g_i]        
#         problem += U_Cm <= CORES
#         
#         problem += objVar >= U_Cm
#     
#     problem += objVar
#     
#     # print problem 
#     start = time.clock() 
# #     print "solving"
#     status = problem.solve(GUROBI(msg=0, timeLimit = 480))
# #     print "solved"
#     fin = time.clock()
#     
#     rntm = -1
#     if LpStatus[status] != 'Not Solved':
#         rntm = fin-start
# 
#     if LpStatus[status] == 'Optimal':
# #         print "PAGES PER SLOT:"
# #         for b_j in range(c0, 4):
# #             print "  CORE " + str(b_j) + ":"
# #             print "    pp      :", np.sum([np.sum([use[g_i, b_j, c_k].varValue*pp[g_i, b_j, c_k].varValue for g_i in range(mdn)]) for c_k in range(16)])
# #             tsAB = ts.level('A')
# #             tsAB.extend(ts.level('B'))
# #             mpt = tsAB.core(b_j).num_pages()
# #             print "    act_pp  :", mpt
# #         if c0 == 1:
# #             print " SHARED: " 
# #             print "    colorsSH  :", colorsSH.varValue
# #         for g_i in range(mdn):
# #             print "MODE", g_i
# #             print " UTIL:"
# #             print "  A:"
# #             print "    CORES: " 
# #             print "      A:", [UA_A[g_i, b_j].varValue for b_j in range(4)]
# #             print "     dA:", [UA_Adata[g_i][b_j][int(colors[g_i][b_j-c0].value())] < UA_A[g_i, b_j].varValue+.0001 for b_j in range(4)]
# #             print "      B:", [UA_B[g_i, b_j].varValue for b_j in range(4)]
# #             print "     dB:", [UA_Bdata[g_i][b_j][int(colors[g_i][b_j-c0].value())] < UA_B[g_i, b_j].varValue+.0001 for b_j in range(4)]
# #             print "      C:", [UA_C[g_i, b_j].varValue for b_j in range(4)]
# #             print "     dC:", [UA_Cdata[g_i][b_j][int(colors[g_i][b_j-c0].value())] < UA_C[g_i, b_j].varValue+.0001 for b_j in range(4)]
# #             print "    TOTAL: "
# #             print "      A:", np.sum([UA_A[g_i, b_j].varValue for b_j in range(4)])
# #             print "      B:", np.sum([UA_B[g_i, b_j].varValue for b_j in range(4)])
# #             print "      C:", np.sum([UA_C[g_i, b_j].varValue for b_j in range(4)])
# #             print "  B:"
# #             print "    CORES: " 
# #             print "      B:", [UB_B[g_i, b_j].varValue for b_j in range(4)]
# #             print "     dB:", [UB_Bdata[g_i][b_j][int(colors[g_i][b_j-c0].value())] < UB_B[g_i, b_j].varValue+.0001 for b_j in range(4)]
# #             print "      C:", [UB_C[g_i, b_j].varValue for b_j in range(4)]
# #             print "     dC:", [UB_Cdata[g_i][b_j][int(colors[g_i][b_j-c0].value())] < UB_C[g_i, b_j].varValue+.0001 for b_j in range(4)]
# #             print "    TOTAL: "
# #             print "      B:", np.sum([UB_B[g_i, b_j].varValue for b_j in range(4)])
# #             print "      C:", np.sum([UB_C[g_i, b_j].varValue for b_j in range(4)])
# #             print "  C:"
# #             print "    TOTAL: "
# #             print "      C:", UC_Cdata[g_i]
# #             print
# #             
# #             print " PAGES:"
# #             print " DISJOINT:"
# #             for b_j in range(c0, 4):
# #                 print "  CORE " + str(b_j) + ":"
# #                 print "    pp      :", [use[g_i, b_j, c_k].varValue*pp[g_i, b_j, c_k].varValue for c_k in range(16)]
# #                 print "    use     :", [use[g_i, b_j, c_k].varValue for c_k in range(16)]
# #                 print "    pc      :", pc[g_i, b_j].varValue
# #                 print "    colors  :", colors[g_i][b_j-c0].value()
# #                 gtsAB = ts.modes[g_i].level('A')
# #                 gtsAB.extend(ts.modes[g_i].level('B'))
# #                 mpt = gtsAB.core(b_j).num_pages()
# #                 print "    act_pp  :", mpt
# #             if c0 == 1:
# #                 print " SHARED: " 
# #                 print "    colorsSH  :", colorsSH.varValue
# #             print
# #         
# #         for g_i in range(mdn):
# #             print "U_C:", np.sum([UA_C[g_i, b_j].varValue + UB_C[g_i, b_j].varValue for b_j in range(4)])+UC_Cdata[g_i]
# #         print "OBJECTIVE: ", objVar.varValue
# #         print
# #         print
#             
#         objVar = LpVariable("OBJ", lowBound = 0, cat='Continuous')
#         
#         ps = scheme([area(csch.A[i].W, 0) for i in range(0,CORES)],
#                     [area(csch.B[i].W, 0) for i in range(0,CORES)],
#                     csch.C, MAX_WAYS, MAX_SETS)
#         
#         ps.A = [val.rounded() for val in ps.A]
#         ps.B = [val.rounded() for val in ps.B]
#         ps.C = ps.C.rounded()
#         
#         return ps, rntm
#     if LpStatus[status] == 'Not Solved':
#         return "TimeOut", rntm
#         
#     return None, rntm